import greenfoot.*;  // (World, Actor, GreenfootImage, Greenfoot and MouseInfo)
import java.util.*;


public abstract class AbstractTower extends CompositeMapActor {
    private String name;
    private int level;
    private long lastFireTimeMillis = System.currentTimeMillis();
    private TowerRadiusVisualizer radiusVisualizer;
    private Label levelLabel;
    
    public AbstractTower(String name) {
        this.name = name;
        this.levelLabel = new Label(level, CybervirusHazard.getFontName(), 8);
        setLevel(1);
        Toast.show("tower-is-created", "You have bought a " + name + "!", 2000);
        SoundLoader.play("map.ui.tower.buy");
        this.radiusVisualizer = new TowerRadiusVisualizer(this);
    }
    
    public void setLevel(int level) {
        if(level <= 0 || TowerParametersManager.isMaxLevel(name, level)) return;
        this.level = level;
        this.levelLabel.setText(level);
        this.setImage(ImageLoader.get("map.sprite.tower." + name.toLowerCase().replace(" ", "") + "." + level));
        setZoom(2);
    }
    
    public int getLevel() {
        return level;
    }
    
    public void upgrade() {
        setLevel(level + 1);
    }
    
    public abstract void fire(AbstractVirus virus);
    
    public abstract boolean activate();
    
    public void act() {
        super.act();
        List<AbstractVirus> virusesInRadius = getVirusesInRadius();
        if(System.currentTimeMillis() - lastFireTimeMillis >= getParameters().getDelay()) {
            if(activate()) this.lastFireTimeMillis = System.currentTimeMillis();
            if(virusesInRadius.size() != 0) {
                fire(virusesInRadius.get(new Random().nextInt(virusesInRadius.size())));
                this.lastFireTimeMillis = System.currentTimeMillis();
                SoundLoader.play("map.sprite.bullet." + name);
            }
        }
        
        if(Greenfoot.mouseClicked(this)) {
            if(Greenfoot.isKeyDown("control")) {
                int energyFromSelling = 0;
                for(int levelPassed = 1; levelPassed <= level; levelPassed++) {
                    energyFromSelling += (int) Math.ceil(TowerParametersManager.getUpgradePrice(name, levelPassed) * .4);
                }
                new TowerExplosionEffect(this);
                CybervirusHazard.getCurrentMapView().addEnergy(energyFromSelling); // 40% of tower price
                CybervirusHazard.getCurrentMapView().add(new TowerSlot(), getLocation());
                SoundLoader.play("map.ui.tower.sell");
                ParticleSystem.create(getLocation(), 8, "tower-splinter", 6000L, 8000L, .3, -5.5, 5.5, -7.5, 0.5);
                Toast.show("tower-is-sold", "You have sold the " + name + " for " + energyFromSelling + " energy points", 2000, Colors.VERY_LIGHT_SEA);
                suicide();
            } else if(!TowerParametersManager.isMaxLevel(name, level + 1)) {
                if(CybervirusHazard.getCurrentMapView().hasEnoughEnergy(TowerParametersManager.getUpgradePrice(name, level + 1))) {
                    CybervirusHazard.getCurrentMapView().removeEnergy(TowerParametersManager.getUpgradePrice(name, level + 1));
                    upgrade();
                    SoundLoader.play("map.ui.tower.upgrade");
                    ParticleSystem.create(getLocation(), 8, "tower-buy-sparkle", 6000L, 8000L, .1, -2.5, 2.5, -3.5, 0.5, 0.9);
                    Toast.show("tower-is-upgraded", name + " is successfully upgraded to level " + level, 2000);
                } else {
                    Toast.show("lack-of-energy", "You have no enough energy to upgrade this!", 2000, Colors.VERY_LIGHT_PURPLE);
                }
            } else {
                SoundLoader.play("map.ui.tower.maxed");
                Toast.show("tower-is-maxed", "This " + name + " is maxed!", 3000, Colors.VERY_LIGHT_PURPLE);
            }
        }
    }
    
    public TowerParameters getParameters() {
        return TowerParametersManager.get(name, level);
    }
    
    public List<AbstractVirus> getVirusesInRadius() {
        return getObjectsInRange(getParameters().getRadius(), AbstractVirus.class);
    }
    
    public String getName() { return name; }
    
    public TowerRadiusVisualizer getTowerRadiusVisualizer() { return radiusVisualizer; }
    
    public Iterable<CompanionActorHolder> getCompanions() {
        return List.of(
            new CompanionActorHolder(radiusVisualizer, new Location(0, 0)),
            new CompanionActorHolder(levelLabel, new Location(18, 12))
        );
    }
}
